@@ -16,6 +16,7 @@ import functools
 import requests
 import configobj
 import re
+import math
 
 import gi
 gi.require_version('Gtk', '3.0')
@@ -23,12 +24,18 @@ gi.require_version('AppStream', '1.0')
 gi.require_version('XApp', '1.0')
 from gi.repository import Gtk, Gdk, GdkPixbuf, GObject, GLib, Gio, XApp
 
-from installer import installer
-from misc import print_timing
+from mintcommon.installer import installer
+from mintcommon.installer.misc import print_timing
 import reviews
 import housekeeping
 
-ICON_SIZE = 48
+LIST_ICON_SIZE = 48
+FEATURED_ICON_SIZE = LIST_ICON_SIZE
+DETAILS_ICON_SIZE = 64
+SCREEN_THUMB_WIDTH = 100
+SCREENSHOT_WIDTH = 625
+
+FALLBACK_PACKAGE_ICON_PATH = "/usr/share/linuxmint/mintinstall/data/available.png"
 
 #Hardcoded mouse back button key for button-press-event
 #May not work on all mice
@@ -79,6 +86,106 @@ ALIASES['pia-manager'] = "PIA Manager"
 ALIASES['skypeforlinux'] = "Skype"
 ALIASES['google-earth-pro-stable'] = "Google Earth"
 ALIASES['whatsapp-desktop'] = "WhatsApp"
+ALIASES['wine-installer'] = "Wine"
+
+class AsyncImage(Gtk.Image):
+    def __init__(self, icon_string=None, width=DETAILS_ICON_SIZE, height=DETAILS_ICON_SIZE):
+        super(AsyncImage, self).__init__()
+
+        self.path = None
+        self.cancellable = None
+        self.loader = None
+
+        self.connect("destroy", self.on_destroyed)
+
+        if icon_string:
+            self.set_icon_string(icon_string, width, height)
+
+    def on_destroyed(self, widget, data=None):
+        if self.cancellable:
+            self.cancellable.cancel()
+
+    def set_icon_string(self, icon_string, width=DETAILS_ICON_SIZE, height=DETAILS_ICON_SIZE):
+        theme = Gtk.IconTheme.get_default()
+
+        self.original_width = width
+        self.original_height = height
+
+        # This keeps the icon's space occupied until loaded.
+        self.set_size_request(width, height)
+
+        if width != -1:
+            self.width = width * self.get_scale_factor()
+        else:
+            self.width = width
+
+        if height != -1:
+            self.height = height * self.get_scale_factor()
+        else:
+            self.height = height
+
+        self.cancellable = None
+        file = None
+
+        if os.path.isabs(icon_string):
+            self.path = icon_string
+            file = Gio.File.new_for_path(self.path)
+        elif icon_string.startswith("http"):
+            self.path = icon_string
+            file = Gio.File.new_for_uri(self.path)
+        elif theme.has_icon(icon_string):
+                info = theme.lookup_icon_for_scale(icon_string,
+                                                   self.height,
+                                                   self.get_scale_factor(),
+                                                   Gtk.IconLookupFlags.FORCE_SIZE)
+                if info:
+                    self.path = info.get_filename()
+                    file = Gio.File.new_for_path(self.path)
+
+        if file:
+            self.cancellable = Gio.Cancellable()
+            file.read_async(GLib.PRIORITY_DEFAULT, self.cancellable, self.on_read_finished)
+        else:
+            self.set_icon_string(FALLBACK_PACKAGE_ICON_PATH, self.original_width, self.original_height)
+
+    def on_read_finished(self, file, result, data=None):
+        if self.cancellable.is_cancelled():
+            return
+
+        stream = None
+
+        try:
+            stream = file.read_finish(result)
+        except GLib.Error as e:
+            print("AsyncIcon could not read icon file contents for loading (%s): %s" % (self.path, e.message))
+            self.set_icon_string(FALLBACK_PACKAGE_ICON_PATH, self.original_width, self.original_height)
+
+        if stream:
+            GdkPixbuf.Pixbuf.new_from_stream_at_scale_async(stream,
+                                                            self.width,
+                                                            self.height,
+                                                            True,
+                                                            self.cancellable,
+                                                            self.on_pixbuf_created)
+
+    def on_pixbuf_created(self, stream, result, data=None):
+        if self.cancellable.is_cancelled():
+            stream.close()
+            return
+
+        try:
+            pixbuf = GdkPixbuf.Pixbuf.new_from_stream_finish(result)
+
+            if pixbuf:
+                surface = Gdk.cairo_surface_create_from_pixbuf(pixbuf,
+                                                               self.get_scale_factor(),
+                                                               self.get_window())
+                self.set_from_surface(surface)
+        except GLib.Error as e:
+            print("AsyncIcon could not generate icon from file contents (%s): %s" % (self.path, e.message))
+            self.set_icon_string(FALLBACK_PACKAGE_ICON_PATH, self.original_width, self.original_height)
+
+        stream.close()
 
 class ScreenshotDownloader(threading.Thread):
     def __init__(self, application, pkginfo):
@@ -419,6 +526,7 @@ class Application(Gtk.Application):
     PAGE_LIST = "list"
     PAGE_PACKAGE = "details"
     PAGE_LOADING = "loading"
+    PAGE_SEARCHING = "searching"
 
     def __init__(self):
         super(Application, self).__init__(application_id='com.linuxmint.mintinstall',
@@ -585,30 +693,20 @@ class Application(Gtk.Application):
 
     def get_low_res_screen(self):
         display = Gdk.Display.get_default()
-        n_monitors = display.get_n_monitors()
+        screen = display.get_default_screen()
+        pointer = display.get_default_seat().get_pointer()
+        if pointer:
+            position = pointer.get_position()
+            monitor_number = screen.get_monitor_at_point(position.x, position.y)
+        else:
+            monitor_number = 0
+        monitor = display.get_monitor(monitor_number)
+        height = monitor.get_geometry().height
 
-        # Find the shortest monitor height.
-
-        min_height = 99999
-
-        for i in range(n_monitors):
-            monitor = display.get_monitor(i)
-
-            # Look only at geometry, not workarea - this we're looking at the monitor itself,
-            # not affected by someone's panel configuration.
-            rect = monitor.get_geometry()
-
-            min_height = rect.height if rect.height < min_height else min_height
-
-        # If it's less than our threshold than consider us 'low res'.  It's possible we
-        # have false positives here, especially with multi-monitor setups.  We'll always
-        # go off the lowest res monitor, but at the point this is called, we don't know
-        # which monitor we'll spawn on.  Knowing this would require waiting for window
-        # realize, and this really isn't that big a deal.  The workarea being used is in
+        # If it's less than our threshold than consider us 'low res'. The workarea being used is in
         # app pixels, so hidpi will also be affected here regardless of device resolution.
-
-        if min_height < 1000:
-            print("MintInstall: low resolution monitor detected, limiting window height.")
+        if height < 768:
+            print("MintInstall: low resolution detected on monitor %d (%dpx height), limiting window height." % (monitor_number, height))
             return True
 
         return False
@@ -632,6 +730,25 @@ class Application(Gtk.Application):
         self.main_window.connect("key-press-event", self.on_keypress)
         self.main_window.connect("button-press-event", self.on_buttonpress)
 
+        theme = Gtk.IconTheme.get_default()
+        for icon_name in ["application-x-deb", "file-roller"]:
+            if theme.has_icon(icon_name):
+                iconInfo = theme.lookup_icon_for_scale(icon_name,
+                                                       LIST_ICON_SIZE,
+                                                       self.main_window.get_scale_factor(),
+                                                       0)
+                if iconInfo and os.path.exists(iconInfo.get_filename()):
+                    global FALLBACK_PACKAGE_ICON_PATH
+                    FALLBACK_PACKAGE_ICON_PATH = iconInfo.get_filename()
+                    break
+
+        self.detail_view_icon = AsyncImage()
+        self.detail_view_icon.show()
+        self.builder.get_object("application_icon_holder").add(self.detail_view_icon)
+
+        self.main_screenshot = AsyncImage()
+        self.builder.get_object("box_main_screenshot").add(self.main_screenshot)
+
         self.status_label = self.builder.get_object("label_ongoing")
         self.progressbar = self.builder.get_object("progressbar1")
         self.progress_box = self.builder.get_object("progress_box")
@@ -640,7 +757,7 @@ class Application(Gtk.Application):
         self.active_tasks_button = self.builder.get_object("active_tasks_button")
         self.active_tasks_spinner = self.builder.get_object("active_tasks_spinner")
         self.no_packages_found_label = self.builder.get_object("no_packages_found_label")
-        
+
         self.no_packages_found_refresh_button = self.builder.get_object("no_packages_found_refresh_button")
         self.no_packages_found_refresh_button.connect("clicked", self.on_refresh_cache_clicked)
 
@@ -740,18 +857,6 @@ class Application(Gtk.Application):
 
         self.app_list_stack = self.builder.get_object("app_list_stack")
 
-        self.generic_available_icon_path = "/usr/share/linuxmint/mintinstall/data/available.png"
-        theme = Gtk.IconTheme.get_default()
-        for icon_name in ["application-x-deb", "file-roller"]:
-            if theme.has_icon(icon_name):
-                iconInfo = theme.lookup_icon(icon_name, ICON_SIZE, 0)
-                if iconInfo and os.path.exists(iconInfo.get_filename()):
-                    self.generic_available_icon_path = iconInfo.get_filename()
-                    break
-
-        self.generic_available_icon_pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(self.generic_available_icon_path,
-                                                                                    ICON_SIZE, ICON_SIZE)
-
         self.searchentry.grab_focus()
 
         self.listbox_categories = Gtk.ListBox()
@@ -788,7 +893,7 @@ class Application(Gtk.Application):
 
             self.sync_installed_apps()
             self.update_conditional_widgets()
-            
+
             GObject.idle_add(self.finished_loading_packages)
 
             # Can take some time, don't block for it (these are categorizing packages based on apt info, not our listings)
@@ -796,7 +901,8 @@ class Application(Gtk.Application):
 
             self.review_cache = reviews.ReviewCache()
             housekeeping.run()
-        except:
+        except GLib.Error as e:
+            print("Loading error: %s", e.message)
             GObject.idle_add(self.refresh_cache)
 
     def load_featured_on_landing(self):
@@ -893,8 +999,7 @@ class Application(Gtk.Application):
         random.shuffle(available)
         featured = 0
         for pkginfo in (available + installed):
-            icon = self.get_application_icon(pkginfo, ICON_SIZE)
-            icon = Gtk.Image.new_from_pixbuf(icon)
+            icon = self.get_application_icon(pkginfo, FEATURED_ICON_SIZE)
             tile = VerticalPackageTile(pkginfo, icon, self.installer)
             tile.connect("clicked", self.on_flowbox_item_clicked, pkginfo.pkg_hash)
             flowbox.insert(tile, -1)
@@ -1045,10 +1150,12 @@ class Application(Gtk.Application):
         return return_list
 
     def sync_installed_apps(self):
-        # garbage collect any stale packages in this list (uninstalled somewhere else)
-        installed_packages = self.settings.get_strv(INSTALLED_APPS)
+        """ garbage collect any stale packages in this list (uninstalled somewhere else) """
 
-        installed_packages = self.modernize_installed_list(installed_packages)
+        installed_packages = self.installer.cache.get_manually_installed_packages()
+        if not installed_packages:
+            installed_packages = self.settings.get_strv(INSTALLED_APPS)
+            installed_packages = self.modernize_installed_list(installed_packages)
 
         for pkg_hash in installed_packages:
             try:
@@ -1085,11 +1192,10 @@ class Application(Gtk.Application):
             if (number == 1):
                 if os.path.exists(local_name):
                     try:
-                        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(local_name, 625, -1)
-                        self.builder.get_object("main_screenshot").set_from_pixbuf(pixbuf)
-                        self.builder.get_object("main_screenshot").show()
+                        self.main_screenshot.set_icon_string(local_name, SCREENSHOT_WIDTH, -1)
+                        self.main_screenshot.show()
                     except Exception:
-                        self.builder.get_object("main_screenshot").hide()
+                        self.main_screenshot.hide()
                         print("Invalid picture %s, deleting." % local_name)
                         os.unlink(local_name)
                         os.unlink(local_thumb)
@@ -1099,11 +1205,10 @@ class Application(Gtk.Application):
                         try:
                             name = os.path.join(SCREENSHOT_DIR, "%s_1.png" % pkg_name)
                             thumb = os.path.join(SCREENSHOT_DIR, "thumb_%s_1.png" % pkg_name)
-                            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(thumb, 100, -1)
                             event_box = Gtk.EventBox()
-                            image = Gtk.Image.new_from_pixbuf(pixbuf)
+                            image = AsyncImage(thumb, SCREEN_THUMB_WIDTH, -1)
                             event_box.add(image)
-                            event_box.connect("button-release-event", 
+                            event_box.connect("button-release-event",
                                               self.on_screenshot_clicked,
                                               image,
                                               thumb,
@@ -1113,9 +1218,8 @@ class Application(Gtk.Application):
                         except Exception:
                             pass
                     try:
-                        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(local_thumb, 100, -1)
                         event_box = Gtk.EventBox()
-                        image = Gtk.Image.new_from_pixbuf(pixbuf)
+                        image = AsyncImage(local_thumb, SCREEN_THUMB_WIDTH, -1)
                         event_box.add(image)
                         event_box.connect("button-release-event",
                                           self.on_screenshot_clicked,
@@ -1131,8 +1235,7 @@ class Application(Gtk.Application):
 
     def on_screenshot_clicked(self, eventbox, event, image, local_thumb, local_name):
         # Set main screenshot
-        pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(local_name, 625, -1)
-        self.builder.get_object("main_screenshot").set_from_pixbuf(pixbuf)
+        self.main_screenshot.set_icon_string(local_name, SCREENSHOT_WIDTH, -1)
 
     def category_button_clicked(self, button, category):
         self.show_category(category)
@@ -1152,7 +1255,7 @@ class Application(Gtk.Application):
         terms = searchentry.get_text()
 
         if terms != "":
-            self.show_search_results(terms);
+            self.show_search_results(terms)
 
     def on_entry_text_changed(self, entry):
         if self.search_changed_timer > 0:
@@ -1196,12 +1299,8 @@ class Application(Gtk.Application):
             dlg.set_license(gpl)
         except Exception as e:
             print(e)
-        try:
-            version = subprocess.check_output(["/usr/lib/linuxmint/common/version.py", "mintinstall"]).decode()
-            dlg.set_version(version)
-        except Exception as e:
-            print(e)
 
+        dlg.set_version("__DEB_VERSION__")
         dlg.set_icon_name("mintinstall")
         dlg.set_logo_icon_name("mintinstall")
 
@@ -1281,7 +1380,8 @@ class Application(Gtk.Application):
 
         # kill -9 won't kill mp subprocesses, we have to do them ourselves.
         housekeeping.kill()
-        self.review_cache.kill()
+        if self.review_cache:
+            self.review_cache.kill()
 
         # Not happy with Python when it comes to closing threads, so here's a radical method to get what we want.
         os.system("kill -9 %s &" % os.getpid())
@@ -1687,40 +1787,18 @@ class Application(Gtk.Application):
     def on_row_activated(self, listbox, row):
         self.show_category(row.category)
 
+    def get_application_icon_string(self, pkginfo, size):
+        string = self.installer.get_icon(pkginfo, size)
+
+        if not string:
+            string = FALLBACK_PACKAGE_ICON_PATH
+
+        return string
+
     def get_application_icon(self, pkginfo, size):
-        # Look in the icon theme first
-        theme = Gtk.IconTheme.get_default()
+        icon_string = self.get_application_icon_string(pkginfo, size)
 
-        try:
-            for name in [pkginfo.name, pkginfo.name.split(":")[0], pkginfo.name.split("-")[0], pkginfo.name.split(".")[-1].lower()]:
-                if theme.has_icon(name):
-                    iconInfo = theme.lookup_icon(name, size, 0)
-                    if iconInfo and os.path.exists(iconInfo.get_filename()):
-                        return GdkPixbuf.Pixbuf.new_from_file_at_size(iconInfo.get_filename(), size, size)
-
-            # For Flatpak, look in Appstream and mintinstall provided flatpak icons
-            if pkginfo.pkg_hash.startswith("f"):
-                icon_path = "/var/lib/flatpak/appstream/%s/%s/active/icons/64x64/%s.png" % (pkginfo.remote, pkginfo.arch, pkginfo.name)
-                if os.path.exists(icon_path):
-                    return GdkPixbuf.Pixbuf.new_from_file_at_size(icon_path, size, size)
-
-                icon_path = "/usr/share/linuxmint/mintinstall/flatpak/icons/64x64/%s.png" % pkginfo.name
-                if os.path.exists(icon_path):
-                    return GdkPixbuf.Pixbuf.new_from_file_at_size(icon_path, size, size)
-
-            # Look in app-install-data and pixmaps
-            for extension in ['svg', 'png', 'xpm']:
-                icon_path = "/usr/share/app-install/icons/%s.%s" % (pkginfo.name, extension)
-                if os.path.exists(icon_path):
-                    return GdkPixbuf.Pixbuf.new_from_file_at_size(icon_path, size, size)
-
-                icon_path = "/usr/share/pixmaps/%s.%s" % (pkginfo.name, extension)
-                if os.path.exists(icon_path):
-                    return GdkPixbuf.Pixbuf.new_from_file_at_size(icon_path, size, size)
-        except GLib.Error:
-            pass
-
-        return self.generic_available_icon_pixbuf
+        return AsyncImage(icon_string, size, size)
 
     @print_timing
     def show_search_results(self, terms):
@@ -1745,7 +1823,9 @@ class Application(Gtk.Application):
         self.listbox_categories.hide()
         self.back_button.set_sensitive(True)
         self.previous_page = self.PAGE_LANDING
-        self.page_stack.set_visible_child_name(self.PAGE_LIST)
+        if self.page_stack.get_visible_child_name() != self.PAGE_SEARCHING:
+            self.builder.get_object("loading_spinner").start()
+            self.page_stack.set_visible_child_name(self.PAGE_SEARCHING)
 
         termsUpper = terms.upper()
         termsSplit = re.split(r'\W+', termsUpper)
@@ -1791,6 +1871,8 @@ class Application(Gtk.Application):
 
     def on_search_results_complete(self, results):
         self.clear_category_list()
+        self.page_stack.set_visible_child_name(self.PAGE_LIST)
+        self.builder.get_object("loading_spinner").stop()
         self.show_packages(results, from_search=True)
 
     def on_flowbox_item_clicked(self, tile, data=None):
@@ -1854,7 +1936,7 @@ class Application(Gtk.Application):
                 else:
                     text = _("No packages to show.\nThis may indicate a problem - try refreshing the cache.")
                     self.no_packages_found_refresh_button.show()
-                    
+
             self.no_packages_found_label.set_markup("<big><b>%s</b></big>" % text)
         else:
             self.app_list_stack.set_visible_child_name("results")
@@ -1920,8 +2002,7 @@ class Application(Gtk.Application):
             self.one_package_idle_timer = 0
             return False
 
-        icon = self.get_application_icon(pkginfo, ICON_SIZE)
-        icon = Gtk.Image.new_from_pixbuf(icon)
+        icon = self.get_application_icon(pkginfo, LIST_ICON_SIZE)
 
         summary = self.installer.get_summary(pkginfo)
         summary = summary.replace("<", "&lt;")
@@ -1996,7 +2077,8 @@ class Application(Gtk.Application):
 
         self.builder.get_object("label_details").set_markup(details_markup)
 
-        self.builder.get_object("application_icon").set_from_pixbuf(self.get_application_icon(pkginfo, 64))
+        icon_string = self.get_application_icon_string(pkginfo, DETAILS_ICON_SIZE)
+        self.detail_view_icon.set_icon_string(icon_string)
         self.builder.get_object("application_name").set_label(self.installer.get_display_name(pkginfo))
         self.builder.get_object("application_summary").set_label(self.installer.get_summary(pkginfo))
         self.builder.get_object("application_package").set_label(pkginfo.name)
@@ -2080,16 +2162,15 @@ class Application(Gtk.Application):
         main_thumb = os.path.join(SCREENSHOT_DIR, "thumb_%s_1.png" % pkginfo.name)
         if os.path.exists(main_screenshot) and os.path.exists(main_thumb):
             try:
-                main_screenshot = GdkPixbuf.Pixbuf.new_from_file_at_size(main_screenshot, 625, -1)
-                self.builder.get_object("main_screenshot").set_from_pixbuf(main_screenshot)
-                self.builder.get_object("main_screenshot").show()
+                self.main_screenshot.set_icon_string(main_screenshot, SCREENSHOT_WIDTH, -1)
+                self.main_screenshot.show()
             except:
-                self.builder.get_object("main_screenshot").hide()
+                self.main_screenshot.hide()
                 os.unlink(main_screenshot)
             for i in range(2, 5):
                 self.add_screenshot(pkginfo.name, i)
         else:
-            self.builder.get_object("main_screenshot").hide()
+            self.main_screenshot.hide()
             downloadScreenshots = ScreenshotDownloader(self, pkginfo)
             downloadScreenshots.start()
 
@@ -2123,6 +2204,8 @@ class Application(Gtk.Application):
         # Load package info
         style_context = self.action_button.get_style_context()
 
+        action_button_description = ""
+
         if task.info_ready_status == task.STATUS_OK:
             if task.type == "remove":
                 action_button_label = _("Remove")
